use color_eyre::eyre::{Result, WrapErr};
use crossterm::event::KeyEvent;
use nonempty::NonEmpty;
use ratatui::prelude::Rect;
use serde::{Deserialize, Serialize};
use tokio::sync::mpsc;

use crate::{
  action::Action,
  commands::RataCommands,
  components::{fps::FpsCounter, home::Home, Component},
  config::Config,
  mode::Mode,
  tui,
};

pub struct App {
  pub config: Config,
  pub clients: NonEmpty<matrix_sdk::Client>,
  pub commands: RataCommands,
  pub tick_rate: f64,
  pub frame_rate: f64,
  pub components: Vec<Box<dyn Component>>,
  pub should_quit: bool,
  pub should_suspend: bool,
  pub mode: Mode,
  pub last_tick_key_events: Vec<KeyEvent>,
}

impl App {
  pub async fn new(tick_rate: f64, frame_rate: f64) -> Result<Self> {
    let home = Home::new();
    let fps = FpsCounter::default();
    let config = Config::new()?;
    let mode = Mode::Home;
    let future_clients = config.accounts.clone().map(|conf| {
      tokio::spawn(async move {
        let server_name = conf.user_id.server_name();
        let client = matrix_sdk::Client::builder()
          .server_name(server_name)
          .build()
          .await
          .with_context(|| format!("Could not initialize client for {}", server_name))?;
        client
          .login_username(&conf.user_id, &conf.password)
          .send()
          .await
          .with_context(|| format!("Could not login as {}", conf.user_id))?;
        Ok::<_, color_eyre::eyre::Report>(client)
      })
    });
    let mut clients = Vec::new();
    for client in futures::future::join_all(future_clients).await {
      clients.push(
        client
          .context("Could not join client init task")?
          .context("Could not connect to all accounts")?,
      );
    }
    let clients = NonEmpty::collect(clients).expect("map on NonEmpty returned empty vec");

    let mut app = Self {
      tick_rate,
      frame_rate,
      components: vec![Box::new(home), Box::new(fps)],
      should_quit: false,
      should_suspend: false,
      clients,
      config,
      commands: RataCommands::new(),
      mode,
      last_tick_key_events: Vec::new(),
    };
    crate::plugins::load_all(&mut app).context("Could not load plugins")?;
    Ok(app)
  }

  pub async fn run(&mut self) -> Result<()> {
    let (action_tx, mut action_rx) = mpsc::unbounded_channel();

    let mut tui = tui::Tui::new()?
      .tick_rate(self.tick_rate)
      .frame_rate(self.frame_rate);
    // tui.mouse(true);
    tui.enter()?;

    for component in self.components.iter_mut() {
      component.register_action_handler(action_tx.clone())?;
    }

    for component in self.components.iter_mut() {
      component.register_config_handler(self.config.clone())?;
    }

    for component in self.components.iter_mut() {
      component.init(tui.size()?)?;
    }

    loop {
      if let Some(e) = tui.next().await {
        match e {
          tui::Event::Quit => action_tx.send(Action::Quit)?,
          tui::Event::Tick => action_tx.send(Action::Tick)?,
          tui::Event::Render => action_tx.send(Action::Render)?,
          tui::Event::Resize(x, y) => action_tx.send(Action::Resize(x, y))?,
          tui::Event::Key(key) => {
            if let Some(keymap) = self.config.keybindings.get(&self.mode) {
              if let Some(command_line) = keymap.get(&vec![key]) {
                log::info!("Got command: {command_line}");
                crate::commands::run_command(command_line, &self, &action_tx)?;
              } else {
                // If the key was not handled as a single key action,
                // then consider it for multi-key combinations.
                self.last_tick_key_events.push(key);

                // Check for multi-key combinations
                if let Some(command_line) = keymap.get(&self.last_tick_key_events) {
                  log::info!("Got command: {command_line}");
                  crate::commands::run_command(command_line, &self, &action_tx)?;
                }
              }
            };
          },
          _ => {},
        }
        for component in self.components.iter_mut() {
          if let Some(action) = component.handle_events(Some(e.clone()))? {
            action_tx.send(action)?;
          }
        }
      }

      while let Ok(action) = action_rx.try_recv() {
        if action != Action::Tick && action != Action::Render {
          log::debug!("{action:?}");
        }
        match action {
          Action::Tick => {
            self.last_tick_key_events.drain(..);
          },
          Action::Quit => self.should_quit = true,
          Action::Suspend => self.should_suspend = true,
          Action::Resume => self.should_suspend = false,
          Action::Resize(w, h) => {
            tui.resize(Rect::new(0, 0, w, h))?;
            tui.draw(|f| {
              for component in self.components.iter_mut() {
                let r = component.draw(f, f.size());
                if let Err(e) = r {
                  action_tx
                    .send(Action::Error(format!("Failed to draw: {:?}", e)))
                    .unwrap();
                }
              }
            })?;
          },
          Action::Render => {
            tui.draw(|f| {
              for component in self.components.iter_mut() {
                let r = component.draw(f, f.size());
                if let Err(e) = r {
                  action_tx
                    .send(Action::Error(format!("Failed to draw: {:?}", e)))
                    .unwrap();
                }
              }
            })?;
          },
          _ => {},
        }
        for component in self.components.iter_mut() {
          if let Some(action) = component.update(action.clone())? {
            action_tx.send(action)?
          };
        }
      }
      if self.should_suspend {
        tui.suspend()?;
        action_tx.send(Action::Resume)?;
        tui = tui::Tui::new()?
          .tick_rate(self.tick_rate)
          .frame_rate(self.frame_rate);
        // tui.mouse(true);
        tui.enter()?;
      } else if self.should_quit {
        tui.stop()?;
        break;
      }
    }
    tui.exit()?;
    Ok(())
  }
}
